# Copyright (c) 2024 iiPython

# Modules
from typing import List
from types import FunctionType

from nightwatch import __version__
from nightwatch.config import config

# Main class
class BaseCommand():
    def __init__(self, name: str, ui, add_message: FunctionType) -> None:
        self.name, self.ui = name, ui
        self.add_message = add_message

    def print(self, message: str) -> None:
        self.add_message(self.name.title(), message)

# Commands
class ShrugCommand(BaseCommand):
    def __init__(self, *args) -> None:
        super().__init__("shrug", *args)

    def on_execute(self, args: List[str]) -> str:
        return "¯\_(ツ)_/¯"

class ConfigCommand(BaseCommand):
    def __init__(self, *args) -> None:
        super().__init__("config", *args)

    def on_execute(self, args: List[str]) -> None:
        if not args:
            for line in [
                "Nightwatch client configuration",
                "Usage: /config <key> <value>",
                "",
                "Example usage:",
                "/config colors.time yellow",
                "/config prompt \">> \"",
                "",
                "Some changes will only apply after Nightwatch restarts."
            ]:
                self.print(line)

            return

        elif len(args) < 2:
            return self.print(f"Missing the value to assign to '{args[0]}'.")

        config.set(args[0], args[1])
        self.print(f"{args[0]} has been set to \"{args[1]}\".")

class HelpCommand(BaseCommand):
    def __init__(self, *args) -> None:
        super().__init__("help", *args)

    def on_execute(self, args: List[str]) -> None:
        self.print(f"✨ Nightwatch v{__version__}")
        self.print("Available commands:")
        for command in self.ui.commands:
            self.print(f"  /{command}")

class MembersCommand(BaseCommand):
    def __init__(self, *args) -> None:
        super().__init__("members", *args)

    def on_execute(self, args: List[str]) -> None:
        def members_callback(response: dict):
            self.print(", ".join(response["members"]))

        self.ui.websocket.callback({"type": "members"}, members_callback)

commands = [
    ShrugCommand,
    ConfigCommand,
    HelpCommand,
    MembersCommand
]



# Copyright (c) 2024 iiPython

# Modules
import re
import json

from fastapi import FastAPI, WebSocket, WebSocketDisconnect

from nightwatch import __version__
from nightwatch.config import config

# Initialization
print(f"✨ Nightwatch | v{__version__}")

app = FastAPI(openapi_url = None)

# Constants
server_name = config["server.name"] or "Untitled Server"
hex_regex = re.compile(r"^#(?:[0-9a-fA-F]{3}){1,2}$")

# Manager class
class Manager():
    def __init__(self) -> None:
        self.connections = {}
        self.commands = {
            "message": (self._message, ["text"], []),
            "identify": (self._identify, ["name", "color"], ["color"]),
            "members": (self._members, [], [])
        }
    
    async def broadcast(self, data: dict) -> None:
        for conn in self.connections:
            try:
                await conn.send_json(data)

            except Exception:
                await conn.close()

    async def _message(self, ws: WebSocket, text: str) -> None:
        if ws not in self.connections:
            return await ws.send_json({"text": "You must identify before sending a message."})

        text = str(text)
        if not text.strip():
            return await ws.send_json({"text": "You cannot send an empty message."})

        elif len(text) > 300:
            return await ws.send_json({"text": "Message is too large, maximum limit is 300 characters."})

        user = self.connections[ws]
        await self.broadcast({"user": user, "text": text})

    async def _identify(self, ws: WebSocket, name: str, color: str) -> None:
        color = str(color or "")
        for check in [
            (lambda: color and not re.search(hex_regex, color), "Specified user color is not a valid HEX code."),
            (lambda: not name.strip(), "No username specified."),
            (lambda: name.lower() in ["nightwatch", "admin", "administrator", "moderator"], "Specified username is reserved."),
            (lambda: ws in self.connections, "You have already identified."),
            (lambda: name in [u["name"] for u in self.connections.values()], "Specified username is already taken."),
            (lambda: len(name) > 30, "Specified username is too large. Maximum is 30 characters.")
        ]:
            if check[0]():
                return await ws.send_json({"text": check[1]})

        self.connections[ws] = {"name": name, "color": color or "yellow"}
        await ws.send_json({"online": len(self.connections), "name": server_name})
        await self.broadcast({"text": f"{name} joined the chatroom."})

    async def _members(self, ws: WebSocket) -> None:
        await ws.send_json({"members": [u["name"] for u in self.connections.values()]})

    async def handle_message(self, ws: WebSocket, message: dict) -> None:
        if "type" not in message:
            return await ws.close()

        elif message["type"] not in self.commands:
            return await ws.send_json({"text": "Specified type is invalid."})

        # Handle command
        if "callback" in message:
            setattr(ws, "callback", message["callback"])

        args = []
        callback, arg_list, optional = self.commands[message["type"]]
        for _ in arg_list:
            if _ not in message and _ not in optional:
                return await ws.send_json({"text": f"Missing argument: '{_}'."})

            args.append(message.get(_, None))

        await callback(ws, *args)

manager = Manager()

# Custom websocket wrapper
class Websocket():
    def __init__(self, ws: WebSocket) -> None:
        self.ws, self.callback = ws, None
        self.receive_json = self.ws.receive_json

    async def send_json(self, data: dict) -> None:
        if self.callback is not None:
            data["callback"] = self.callback

        self.callback = None
        return await self.ws.send_json(data)

# Handle websocket routing
@app.websocket("/gateway")
async def gateway_endpoint(websocket: WebSocket) -> None:
    await websocket.accept()

    # Main loop
    try:
        websocket = Websocket(websocket)
        async for message in websocket.ws.iter_json():
            await manager.handle_message(websocket, message)

    except (KeyError, WebSocketDisconnect, json.JSONDecodeError):
        pass

    # Clean up and free username
    if websocket in manager.connections:
        del manager.connections[websocket]



# Copyright (c) 2024 iiPython

# Initialization
def main() -> None:

    # Check urwid
    def invalid_urwid_distrib():
        exit("ERROR: Nightwatch requires urwid to have a specific unicode patch not currently included in upstream urwid.\n\n" \
            "To fix this, run:\npip install -U urwid@git+https://github.com/iiPythonx/urwid\n\n" \
            "and relaunch Nightwatch.")

    try:
        from urwid import is_iipython_urwid
        if not is_iipython_urwid:
            invalid_urwid_distrib()

    except ImportError:
        invalid_urwid_distrib()

    # Modules
    from argparse import ArgumentParser
    from nightwatch.client import start_client

    # Handle CLI options
    ap = ArgumentParser(
        prog = "nightwatch",
        description = "The chatting application to end all chatting applications.",
        epilog = "Copyright (c) 2024 iiPython\nhttps://github.com/iiPythonx/nightwatch"
    )
    ap.add_argument("-a", "--address", help = "the nightwatch server to connect to")
    ap.add_argument("-u", "--username", help = "the username to use")

    # Launch client
    args = ap.parse_args()
    start_client(args.address, args.username)



# Copyright (c) 2024 iiPython

# Modules
import shlex
from typing import Tuple
from datetime import datetime
from types import FunctionType

import urwid

from .commands import commands
from .wswrap import ORJSONWebSocket
from ..vendor.scroll import Scrollable, ScrollBar

# Input edit class
class InputEdit(urwid.Edit):
    def __init__(self, send_message: FunctionType) -> None:
        self.send_message = send_message
        super().__init__("> ", wrap = "clip")

    def keypress(self, size: Tuple[int], key: str) -> None:
        if key != "enter":
            return super().keypress(size, key)

        elif not self.edit_text.strip():
            return

        self.send_message(self.edit_text)
        self.edit_text = ""

# Main UI class
class NightwatchUI():
    def __init__(self, websocket: ORJSONWebSocket) -> None:
        self.last_author, self.last_time = None, 0
        self.websocket = websocket

        # Main urwid setup
        self.pile = urwid.Pile([])
        self.scroll = Scrollable(urwid.Filler(self.pile, valign = "top"))
        self.frame = urwid.Frame(ScrollBar(self.scroll), footer = InputEdit(self.send_message), focus_part = "footer")

        # Initialize commands
        self.commands = {}
        for command in commands:
            command = command(self, self.add_message)
            self.commands[command.name] = command

    def length(self, text: str) -> int:
        return int(len(text.encode(encoding = "utf_16_le")) / 2)

    def send_message(self, text: str) -> None:
        if text[0] == "/":
            splices = shlex.split(text)
            if splices[0][1:] in self.commands:
                response = self.commands[splices[0][1:]].on_execute(splices[1:])
                if response is None:
                    return  # Internal command, do not send in chat lmao

                text = response

        self.websocket.send({"type": "message", "text": text})

    def construct_message(self, author: str, content: str, user_color: str = "gray") -> None:
        visible_author = author if author != self.last_author else " " * self.length(author)
        now, time_string = datetime.now(), ""
        if (author != self.last_author) or ((now - self.last_time).total_seconds() > 300):
            time_string = now.strftime("%I:%M %p") + "  "  # Right padding for the scrollbar

        self.pile.contents.append((urwid.Columns([
            (self.length(visible_author), urwid.Text((user_color, visible_author))),
            (3, urwid.Text(("sep", " | "))),
            ("weight", 4, urwid.Text(content)),
            (len(time_string) + 2, urwid.Text(("time", time_string), align = "right"))  # +2 adds left padding
        ]), self.pile.options()))
        self.last_author, self.last_time = author, now

    def add_message(self, author: str, content: str, user_color: str = "gray") -> None:
        self.construct_message(author, content, user_color)

        # Scroll to the end (if we haven't manually moved)
        if len(self.pile.contents) > 50:
            self.pile.contents = self.pile.contents[1:]

        self.scroll.to_end()
        self.loop.screen.clear()
        self.loop.draw_screen()

    def on_message(self, data: dict) -> None:
        if not hasattr(self, "loop"):
            return  # We aren't even initialized yet

        elif "callback" in data:
            callback = self.websocket.callbacks[data["callback"]]
            del self.websocket.callbacks[data["callback"]]
            return callback(data)

        user, color_code = data.get("user", {"name": "Nightwatch"}), "gray"
        if user["name"] != "Nightwatch":

            # Handle colors and fallbacks
            color_code = f"user-{user['name']}"
            self.loop.screen.register_palette_entry(color_code, "yellow", "", foreground_high = user["color"])

        # Push message to screen
        self.add_message(user["name"], data["text"], color_code)

    def on_ready(self, loop: urwid.MainLoop, payload: dict) -> None:
        self.loop = loop
        self.construct_message("Nightwatch", f"Welcome to {payload['name']}. There are {payload['online']} user(s) online.")



# Copyright (c) 2024 iiPython

# Modules
import os
from threading import Thread

import urwid
import websockets
from websockets.sync.client import connect

from nightwatch import __version__
from nightwatch.config import config

from .extra.ui import NightwatchUI
from .extra.select import menu
from .extra.wswrap import ORJSONWebSocket

# Initialization
if os.name == "nt":
    urwid.set_encoding("utf-8")

# Connection handler
def connect_loop(host: str, port: int) -> None:
    destination = f"ws{'s' if port == 443 else ''}://{host}:{port}/gateway"
    try:
        with connect(destination) as ws:
            ws = ORJSONWebSocket(ws)

            # Handle identification payload
            ws.send({"type": "identify", "name": config["user.name"], "color": config["user.color"]})
            response = ws.recv()
            if "text" in response:
                exit(f"\nCould not connect to {destination}. Additional details:\n{response['text']}")

            # Create UI
            ui = NightwatchUI(ws)
            loop = urwid.MainLoop(ui.frame, [
                ("yellow", "yellow", ""),
                ("gray", "dark gray", "", "", "#555753", ""),
                ("green", "dark green", "")
            ])
            loop.screen.set_terminal_properties(2 ** 24)  # Activate 24-bit color mode

            # Individual components
            loop.screen.register_palette_entry("time", "dark green", "", foreground_high = config["colors.time"] or "#00FF00")
            loop.screen.register_palette_entry("sep", "dark gray", "", foreground_high = config["colors.sep"] or "#555753")

            # Handle messages
            def message_loop(ws: ORJSONWebSocket, ui: NightwatchUI) -> None:
                try:
                    while ws.ws:
                        ui.on_message(ws.recv())

                except websockets.exceptions.ConnectionClosed:
                    return

            Thread(target = message_loop, args = [ws, ui]).start()

            # Start mainloop
            ui.on_ready(loop, response)
            loop.run()

    except websockets.exceptions.InvalidURI:
        exit(f"\nCould not connect to {destination} due to an HTTP redirect.\nPlease ensure you entered the correct address.")

    except OSError:
        exit(f"\nCould not connect to {destination} due to an OSError.\nThis is more then likely because the server is not running.")

# Entrypoint
def start_client(
    address: str = None,
    username: str = None
):
    username = username or config["user.name"]

    # Start main UI
    print(f"\033[H\033[2J✨ Nightwatch | v{__version__}\n")
    if username is None:
        print("Hello! It seems that this is your first time using Nightwatch.")
        print("Before you can connect to a server, please set your desired username.\n")
        config.set("user.name", input("Username: "))
        print("\033[4A\033[0J", end = "")  # Reset back up to the Nightwatch label

    # Handle server address
    if address is None:
        servers = config["servers"]
        if servers is None:
            servers = ["nightwatch.iipython.dev"]
            config.set("servers", servers)

        print(f"Hello, {config['user.name']}. Please select a Nightwatch server to connect to:")
        address = menu.show(servers)
        print()

    print(f"Establishing connection to {address} ...")

    # Parse said address
    if ":" not in address:
        host, port = address, 443

    else:
        host, port = address.split(":")

    # Connect to server
    try:
        connect_loop(host, port)

    except KeyboardInterrupt:
        print("\033[5A\033[0J", end = "")  # Reset back up to the Nightwatch label
        print(f"Goodbye, {config['user.name']}.")



# Copyright (c) 2024 iiPython

# Modules
import os
import json
import typing
from pathlib import Path
from getpass import getuser

# Initialization
config_path = Path(os.path.expanduser("~")) / ".config/nightwatch/config.json"
if os.name == "nt":
    config_path = Path(f"C:\\Users\\{getuser()}\\AppData\\Local\\Nightwatch\\config.json")

config_path.parent.mkdir(exist_ok = True)

# Configuration class
class Configuration():
    def __init__(self, config_path: Path) -> None:
        self.config, self.config_path = {}, config_path
        if config_path.is_file():
            with config_path.open() as fh:
                self.config = json.loads(fh.read())

    def __getitem__(self, item: str) -> typing.Any:
        v = self.config
        for k in item.split("."):
            if k not in v:
                return None

            v = v[k]

        return v

    def set(self, key: str, value: typing.Any) -> None:
        v = self.config
        for k in key.split(".")[:-1]:
            if k not in v:
                v[k] = {}

            v = v[k]

        v[key.split(".")[-1]] = value
        with self.config_path.open("w+") as fh:
            fh.write(json.dumps(self.config, indent = 4))

config = Configuration(config_path)
